import sympy
C = sympy.Symbol('C')
CS = [sympy.Symbol("C_"+str(i)) for i in range(8)]

from lmfit import Parameters


class Model:

	def __init__(self, expr, xs=None, cs=None):
		self.id = -1
		self.parent_id = -1

		self.orig = expr
		self.expr = expr

		self.xs = None
		self.cs = None
		self.params = None

		self.sz = 0

		self.rewrite_coeff()

		params = Parameters()
		for i,c in enumerate(self.cs):
			params.add('C_'+str(i), value=1.0)

		self.params = params

	def __str__(self):
		return str(self.size()) + "  " + str(self.score) + "  " + str(self.expr)


	def get_coeff(self):
		if self.coeff is not None:
			return self.coeff
		else:
			self.coeff = self.rewrite_coeff()

	def size(self):
		if self.sz == 0:
			self.sz = self.calc_tree_size()
		return self.sz

	def calc_tree_size(self):
		i = 0
		for e in sympy.preorder_traversal(self.expr):
			i+=1
		return i


	def rewrite_coeff(self):
		expr, ii = self._rewrite_coeff_helper(self.orig, 0)
		self.expr = expr
		self.cs = CS[:ii]


	def _rewrite_coeff_helper(self, expr, ii):
		ret = expr
		if not expr.is_Atom:
			args = []
			for i,e in enumerate(expr.args):
				if not e.is_Atom:
					ee, ii = self._rewrite_coeff_helper(e,ii)
					args.append(ee)
					# args = args + ee
				elif e == C:
					args.append(CS[ii])
					# args = args + cs[ii]
					ii += 1
				else:
					args.append(e)
			args = tuple(args)
			ret = expr.func(*args)
		return ret,ii



# x,y,z = sympy.symbols("x y z")

# print cs

# f = C*x+C*y+C*z
# F, ii = rewrite_coeff_helper(f,0)
# G, ii = rewrite_coeff_helper(f,3)
# args = (x,y,z)
# F = f.func(*args)
# print F

# g = f.subs(cs)
# print f
# print F
# print G
# print g



from __future__ import division

from sympy import *
init_printing(use_unicode=True)

from itertools import combinations, combinations_with_replacement as combos

BASIC_BASE = (exp, cos, sin)
BASIC_MISC = (Abs, sqrt, log, exp)
BASIC_TRIG = (cos, sin, tan)
HYPER_TRIG = (cosh, sinh, tanh)

C = symbols('C')

class Grower:

	def __init__(self,xs, funcs):
	
		# if only one variable, turn into list
		if type(xs) is Symbol:
			xs = [xs]
	
		self.xs = xs
		self.solo_muls = [ C * x for x in xs ]
		self.double_muls = [ C * tpl[0] * tpl[1] for tpl in combos(xs, 2)]
		self.triple_muls = [ C * tpl[0] * tpl[1] * tpl[2] for tpl in combos(xs, 3)]
		
		self.lin_funcs = []
		self.nonlin_funcs = []
		if funcs is not None:
			self.lin_funcs = [ C*f(x) for f in funcs for x in xs]
			self.nonlin_funcs = [ C*f(C*x+C) for f in funcs for x in xs]

		## these are the same as above, minus the C
		self.var_sub_xs = [ tpl[0] * tpl[1] for tpl in combos(xs, 2) ]
		self.var_sub_fs = []
		if funcs is not None:
			self.var_sub_fs = [ f(x) for f in funcs for x in xs ]
		self.var_sub_terms = self.var_sub_xs + self.var_sub_fs

		self.add_extnd_terms = self.solo_muls + self.double_muls + self.lin_funcs
		self.mul_extnd_terms = self.var_sub_terms

	def first_exprs(self):

		mul_exprs = self.double_muls
		if len(self.xs) < 3:
			mul_exprs += self.triple_muls
		func_exprs = self.lin_funcs

		mid_exprs = self.solo_muls + mul_exprs + func_exprs
		add_exprs = [ Add( tpl[0], tpl[1], evaluate=False ) for tpl in combinations(mid_exprs, 2)]
		plus_C_exprs = [ Add( expr, C ) for expr in mid_exprs + add_exprs]

		ret_exprs = mid_exprs + add_exprs + plus_C_exprs

		return ret_exprs


	def grow(self, model):

		var_expands = self._var_sub(model.orig)
		add_expands = self._add_extend(model.orig)
		mul_expands = self._mul_extend(model.orig)

		return var_expands + add_expands + mul_expands


	def _var_sub(self, expr, limit_sub=False):
		vsubs = []
		# only worry about non-atoms, cause we have to replace args
		if not expr.is_Atom:
			# make a list of args to this non-atom
			# each member of the list is the original expr's args with one substitution made
			args_sets = []
			for i,e in enumerate(expr.args):
				# if the current arg is also a non-atom, recurse
				if not e.is_Atom:
					lim_sub = limit_sub or not (e.is_Add or e.is_Mul)
					# for each expr returned, we need to clone the current args
					# and make the substitution, sorta like flattening?
					ee = self._var_sub(e, lim_sub)
					if len(ee) > 0:
						# We made a substitution(s) on a variable down this branch!!
						for vs in ee:
							# clone current args
							cloned_args = list(expr.args)
							# replace this term in each
							cloned_args[i] = vs
							# append to the args_sets
							args_sets.append(cloned_args)

				elif e in self.xs:
					## Lets make a variable substitution !!
					# loop over self.var_sub_terms
					sub_terms = self.var_sub_terms
					if limit_sub:
						sub_terms = self.var_sub_xs
					for vs in sub_terms:
						# clone current args
						cloned_args = list(expr.args)
						# replace this term in each
						cloned_args[i] = vs
						# append to the args_sets
						args_sets.append(cloned_args)
				else:
					# we don't have to do anything, probably?
					pass

			# finally, create all of the clones at the current level of recursion
			for args in args_sets:
				args = tuple(args)
				tmp = expr.func(*args)
				vsubs.append(tmp)

		return vsubs

		# if e.is_Symbol and e in self.xs:
		

	def _add_extend(self, expr):
		vsubs = []
		# only worry about non-atoms, cause we extend args
		if not expr.is_Atom:
			args_sets = []
			# however we have 2 cases here (as opposed to _var_sub)
			# 1. extend this expression if it's an Add
			if expr.is_Add:
				for term in self.add_extnd_terms:
					cloned_args = list(expr.args)
					cloned_args.append(term)
					args_sets.append(cloned_args)

			# 2. do the recursion
			for i,e in enumerate(expr.args):
				if not e.is_Atom:
					ee = self._add_extend(e)
					if len(ee) > 0:
						# We made a substitution(s) on a variable down this branch!!
						for vs in ee:
							# clone current args
							cloned_args = list(expr.args)
							# replace this term in each
							cloned_args[i] = vs
							# append to the args_sets
							args_sets.append(cloned_args)


			# finally, create all of the clones at the current level of recursion
			for args in args_sets:
				args = tuple(args)
				tmp = expr.func(*args)
				vsubs.append(tmp)

		return vsubs

	def _mul_extend(self, expr):
		vsubs = []
		# only worry about non-atoms, cause we extend args
		if not expr.is_Atom:
			args_sets = []
			# however we have 2 cases here (as opposed to _var_sub)
			# 1. extend this expression if it's an Add
			if expr.is_Mul:
				for term in self.mul_extnd_terms:
					cloned_args = list(expr.args)
					cloned_args.append(term)
					args_sets.append(cloned_args)

			# 2. do the recursion
			for i,e in enumerate(expr.args):
				if not e.is_Atom:
					ee = self._mul_extend(e)
					if len(ee) > 0:
						# We made a substitution(s) on a variable down this branch!!
						for vs in ee:
							# clone current args
							cloned_args = list(expr.args)
							# replace this term in each
							cloned_args[i] = vs
							# append to the args_sets
							args_sets.append(cloned_args)


			# finally, create all of the clones at the current level of recursion
			for args in args_sets:
				args = tuple(args)
				tmp = expr.func(*args)
				vsubs.append(tmp)

		return vsubs








from __future__ import division

from sympy import *
init_printing(use_unicode=True)

import numpy as np
np.random.seed(23)

from sklearn.metrics import mean_squared_error

x = symbols('x')

F_1 = 1.5 * x**2 - x**3
F_2 = exp(Abs(x)) * sin(x)
F_3 = x**2 * exp(sin(x)) + x + sin(pi/4.0 - x**3)

F_1_X = np.linspace(-5., 5., num=200)
F_2_X = np.linspace(-3., 3., num=200)
F_3_X = np.linspace(-10., 10., num=200)

F_1_Y_pure = lambdify(x,F_1,"numpy")(F_1_X)
F_2_Y_pure = lambdify(x,F_2,"numpy")(F_2_X)
F_3_Y_pure = lambdify(x,F_3,"numpy")(F_3_X)

F_1_Y = F_1_Y_pure + np.random.normal(0, 0.05, 200)
F_2_Y = F_2_Y_pure + np.random.normal(0, 0.05, 200)
F_3_Y = F_3_Y_pure + np.random.normal(0, 0.05, 200)

S_1 = mean_squared_error(F_1_Y_pure,F_1_Y)
S_2 = mean_squared_error(F_2_Y_pure,F_2_Y)
S_3 = mean_squared_error(F_3_Y_pure,F_3_Y)

# print S_1, F_1
# print S_2, F_2
# print S_3, F_3




from search import PGE

import expand
import tests


def main():
	print "hello pypge!\n"

	pge_1_var = PGE(
		system_type = "explicit",
		search_vars = "y",
		usable_vars = "x",
		usable_funcs = expand.BASIC_BASE,
		max_iter = 10
		)

	pge_1_var.loop()

	print ""

	# pge_2_var = PGE(
	# 	system_type = "explicit",
	# 	search_vars = "y",
	# 	usable_vars = "x w",
	# 	usable_funcs = expand.BASIC_TRIG,
	# 	max_iter = 10
	# 	)

	# pge.fit( tests.F_2_X, tests.F_2_Y)

main()






class ModelQueue:

	def __init__(self, num_bins=64):
		self.bins = [ [] for i in range(num_bins)]
		self.dirty = [ False for i in range(num_bins)]


	def push(self,model):
		self.bins[model.size()].append(model)
		self.dirty[model.size()] = True
		pass


	def pop(self, cnt):
		# self._sort()
		pops = []
		while len(pops) < cnt:
			last = None

			for i,bin in enumerate( self.bins ):
			# for i,bin in enumerate( reversed(self.bins) ):
			# 	i = 63-i

				# early break
				if len(pops) == cnt:
					return pops

				if len(bin) == 0:
					continue

				if self.dirty[i]:
					bin.sort(key=lambda m: m.score, reverse=True)
					self.dirty[i] = False

				curr = bin[-1]


				if last == None:
					last = curr
					continue
				
				diff = curr.score - last.score
				diff_sz = curr.size() - last.size()

				# print i, diff
				# print "  ", last
				# print "  ", curr

				if diff_sz < 0:
					print "ERROR: size diff should always positive"
					return

				# XXX seems to be an error here
				# should see both 0.002... expressions
				# when solving F_1 test

				# if diff < 0:
					# last = curr
					# continue

				if diff < 0:
					popd = self.bins[last.size()].pop()
					self.dirty[last.size()] = True
					pops.append(popd)

					# print " ++++ ", popd

					if len(pops) + 1 == cnt:
						popd = self.bins[curr.size()].pop()
						self.dirty[curr.size()] = True
						pops.append(popd)
						return pops

					last = curr

			if last is not None:
				popd = self.bins[last.size()].pop()
				self.dirty[last.size()] = True
				pops.append(popd)

	def sort(self):
		for i,bin in enumerate(self.bins):
			self.bins[i].sort(key=lambda m: m.score, reverse=True)
			self.dirty[i] = False


	def _sort(self):
		for i,dty in self.dirty:
			if dty:
				self.bins[i].sort(key=lambda m: m.score, reverse=True)
				self.dirty[i] = False



	def do_print(self):
		offset = 3
		for i,bin in enumerate(self.bins[offset:15]):
			print i+offset, len(bin), self.dirty[i]
			for j,m in enumerate(bin):
				print "   ", j, m.score, m.expr
			print ""
















from __future__ import division
from sympy import *


# print task

eSubs = []
dSubs = []

def buildDeps(deps):

	del eSubs[:]
	del dSubs[:]

	# print deps

	tuples = deps.split('),')
	for t in tuples:
		t = str(t)
		# print "  ", t
		a,b = t.strip().strip('(').strip(')').split(',')
		a,b = a.strip(),b.strip()
		# a, b = str(t[0]), str(t[1])

		sa, sb = symbols(a), symbols(b)
		f = Function(a)(sb)
		df = diff(f,sb,1)

		eSubs.append( [a,b,a+'('+b+')'] )
		dSubs.append( (df,'d'+a+'d'+b) )
		dSubs.append( (f,sa) )

	# print eSubs
	# print dSubs



def do_diff(task):
	eqnStr = task['Eqn'].strip()

	depStr= task['Deps']
	buildDeps(depStr)

	wrts = task['Wrt'].split()

	task["results"]=[]

	for wrt in wrts:
		# replace vars for dependent function version
		tmp = eqnStr
		# print wrt, tmp
		for sub in eSubs:
			if sub[1] == wrt:
				# print sub[0]
				tmp = tmp.replace(sub[0],sub[2])
		# print tmp

		eq = sympify(tmp)


		d_eq = diff(eq, symbols(wrt), 1)
		# print d_eq

		# substitute to look better
		tmp_eq = d_eq
		for sub in dSubs:
			tmp_eq = tmp_eq.subs(sub[0],sub[1])

		task["results"].append(str(tmp_eq))



def do_coeff(task):
	eqnStr = task['Eqn'].strip()
	eq = sympify(eqnStr)

	task["results"]=[]

	coeffs = task['Coeff'].split()

	for c in coeffs:
		d_eq = diff(eq, symbols(c), 1)
		task['result'].append(str(d_eq))



from __future__ import division

import sympy

from lmfit import minimize, Parameters

from sklearn.metrics import mean_squared_error


def Fit(model, xs, X_train, Y_train):
	expr = model.expr

	def fcn2min(params, x_train, y_train):

		c_sub = [ (str(c), params[str(c)].value) for c in model.cs ]
		evlr = expr.subs(c_sub)

		# print evlr

		f = sympy.lambdify(xs, evlr, "numpy")
		y_pred = f(x_train)

		return y_pred - y_train

	result = minimize(fcn2min, model.params, args=(X_train,Y_train))
	model.fit_result = result
	if not result.success:
		print "Error fitting: ", ier, result.message



def Eval(model, xs, X_input):
	c_sub = [ (str(c), model.params[str(c)].value) for c in model.cs ]
	evlr = model.expr.subs(c_sub)

	f = sympy.lambdify(xs, evlr, "numpy")
	y_pred = f(X_input)
	return y_pred

def Score(y_true, y_pred):
	return mean_squared_error(y_true, y_pred)




##  KEEP BELOW HERE AS AN EXAMPLE   ##
##  ALSO FOR THE MIN/MAX ON PARAMS  ##
##  WANT TO EMULATE THE SHIFT FOR TRIG

# from lmfit import minimize, Parameters, report_fit
# import numpy as np

# # create data to be fitted
# x = np.linspace(0, 15, 301)
# data = (5. * np.sin(2 * x - 0.1) * np.exp(-x*x*0.025) +
#         np.random.normal(size=len(x), scale=0.2) )

# # define objective function: returns the array to be minimized
# def fcn2min(params, x, data):
#     """ model decaying sine wave, subtract data"""
#     amp = params['amp'].value
#     shift = params['shift'].value
#     omega = params['omega'].value
#     decay = params['decay'].value

#     model = amp * np.sin(x * omega + shift) * np.exp(-x*x*decay)
#     return model - data

# # create a set of Parameters
# params = Parameters()
# params.add('amp',   value= 10,  min=0)
# params.add('decay', value= 0.1)
# params.add('shift', value= 0.0, min=-np.pi/2., max=np.pi/2)
# params.add('omega', value= 3.0)


# # do fit, here with leastsq model
# result = minimize(fcn2min, params, args=(x, data))

# # calculate final result
# final = data + result.residual

# # write error report
# report_fit(params)

# # try to plot results
# try:
#     import pylab
#     pylab.plot(x, data, 'k+')
#     pylab.plot(x, final, 'r')
#     pylab.show()
# except:
#     pass



import sympy

import tests

import model
import expand
import memoize
import evaluate
import select

import lmfit

class PGE:
	
	def __init__(self,**kwargs):
		# default values
		self.max_iter = 100
		self.pop_count = 3

		self.min_size = 1
		self.max_size = 64
		self.min_depth = 1
		self.max_depth = 6

		self.zero_epsilon = 1e-6
		self.err_method = "mse"

		self.system_type = None
		self.search_vars = None
		self.usable_vars = None
		self.usable_funcs = None

		# override with kwargs
		for key, value in kwargs.items():
			# print key, value
			setattr(self, key, value)


		self.vars = sympy.symbols(self.usable_vars)
		if type(self.vars) is sympy.Symbol:
			self.vars = [self.vars]


		self.prepared = False
		self.prepare()


	def check_config(self):
		if self.system_type == None or self.search_vars == None or self.usable_vars == None:
			return False
		return True


	def prepare(self):

		if not self.check_config():
			print "ERROR: config missing values"
			return

		self.memoizer = memoize.Memoizer(self.vars)
		self.queue = select.ModelQueue(self.max_size)
		self.grower = expand.Grower(self.vars, self.usable_funcs)


		# self.bases = expand.GenerateInitialModels(self.vars, 2, self.usable_funcs)

		for i,e in enumerate(self.grower.first_exprs()):
			m = model.Model(e)
			did_ins = self.memoizer.insert(m)
			size = m.size()
			m.rewrite_coeff()
			# print i,m.expr, size, m.id

			if did_ins:
				evaluate.Fit(m, self.vars, tests.F_1_X, tests.F_1_Y)
				y_pred = evaluate.Eval(m, self.vars, tests.F_1_X)
				m.score = evaluate.Score(tests.F_1_Y, y_pred)
				# print "  ", [ m.params[str(c)].value for c in m.cs ]
				# print "  ", m.score
				self.queue.push(m)

		# self.queue.sort()
		# self.queue.do_print()

		# # print "  pop'n..."
		# popd = self.queue.pop(self.pop_count)
		# print "\npopped:"
		# for p in popd:
		# 	print p

		# # print "  expand..."
		# expanded = []
		# for p in popd:
		# 	ex = self.grower.grow(p)
		# 	expanded.extend(ex)

		# print "\nexpanded:"
		# for e in expanded:
		# 	print e


		self.prepared = True


	def loop(self):
		if not self.prepared:
			print "PGE not prepared"
			return 

		for I in range(self.max_iter):
			print "iter: ", I

			# print "  pop'n..."
			popd = self.queue.pop(self.pop_count)
			print "\npopped:"
			for p in popd:
				print p

			# print "  expand..."
			expanded = []
			for p in popd:
				ex = self.grower.grow(p)
				expanded.extend(ex)

			print "\nexpanded:"
			for e in expanded:
				print e

			for i,e in enumerate(expanded):
				
				try:
					
					# print "  memoize..."
					m = model.Model(e)
					did_ins = self.memoizer.insert(m)
					size = m.size()
					m.rewrite_coeff()

					# print "    if new:"
					if did_ins:
						# print "      train..."
						evaluate.Fit(m, self.vars, tests.F_1_X, tests.F_1_Y)
						if not m.fit_result.success:
							continue
						# print "      score..."
						y_pred = evaluate.Eval(m, self.vars, tests.F_1_X)
						m.score = evaluate.Score(tests.F_1_Y, y_pred)
						# print "      queue..."
						self.queue.push(m)

				except Exception, e:
					print e
					continue




	# sklearn estimator interface functions
	def fit(self, X_train,Y_train):
		self.X_train = X_train
		self.Y_train = Y_train

		print X_train.shape, Y_train.shape
		
		self.loop()

	def predict(self, X_data):
		pass
		











from distutils.core import setup
setup(
  name = 'pypge',
  packages = ['pypge'], # this must be the same as the name above
  version = '0.2',
  description = 'Prioritized Grammar Enumeration implementation',
  author = 'Tony Worm',
  author_email = 'verdverm@gmial.com',
  url = 'https://github.com/verdverm/pypge', # use the URL to the github repo
  download_url = 'https://github.com/verdverm/pypge/tarball/0.1', # I'll explain this in a second
  keywords = ['pge', 'prioritized', 'grammar', 'enumeration', 'symbolic', 'regression', 'genetic programming'], # arbitrary keywords
  classifiers = [],
  install_requires=['sympy', 'pandas', 'lmfit'],
)


from sympy import *
init_printing(use_unicode=True)


class Memoizer:

	def __init__(self, variables):
		self.models = []
		self.mapper = Mapper(variables)
		self.memory = Node()

	def insert(self,model):
		expr = model.orig
		iis, ffs = self.encode(expr)
		return self.insert_encoded(iis,model)

	def insert_encoded(self,iis, model):
		did_ins = self.memory.insert(iis, model)
		if did_ins:
			model.id = len(self.models)
			self.models.append(model)
		return did_ins

	def lookup(self,model):
		expr = model.orig
		iis, ffs = self.encode(expr)
		return self.memory.lookup(iis)

	def lookup_encoded(self,iis):
		return self.memory.lookup(iis)

	def encode(self,expr):
		# print expr
		iis = []
		ffs = []
		for i,e in enumerate(preorder_traversal(expr)):
			# print i, e.func
			ii, ff = self.mapper.get(e)
			iis = iis + ii
			if ff is not None:
				iis.append(len(ffs))
				ffs = ffs + ff
			# print "   ", ii, ff
		return iis, ffs



class Mapper:

	def __init__(self, variables):
		self.variables = variables
		self.coeffs = symbols("C C_0:16")
		self.map = {
			## the Leaf types

			# Symbol: self.map_symbol,
			# Constant 'C_#': 1
			# Float '#.#': 2
			# Time 'T': 3
			# System 'S_#': 4
			# Variable 'X_#': 5
			# Derivative 'dX_#': 6

			numbers.NegativeOne: 7,


			## the Node types
			
			# Mul: 8,  ++ num children
			# Add: 9,  ++ num children
			Pow: 10,
			# Div: 11,

			Abs: 12,
			sqrt: 13,
			log: 14,
			exp: 15,

			cos: 16,
			sin: 17,
			tan: 18,
			
		}
		# print self.variables
		# print self.coeffs

	def get(self,expr):
		e = expr.func
		
		if e is Mul:
			return [8,len(expr.args)], None
		if e is Add:
			return [9,len(expr.args)], None
		
		if e is Symbol:
			return self.map_symbol(expr), None

		if e is Integer or e is numbers.Zero or e is numbers.One:
			return [2, int(expr.evalf(0))], None

		if e is Float or e is numbers.Pi:
			return [2], [expr.evalf(8)]
		
		ii = [self.map[e]]
		return ii, None

	def map_symbol(self,expr):
		# print "SYMBOL: ", expr
		if expr in self.variables:
			idx = self.variables.index(expr)
			return [5,idx]
		if expr in self.coeffs:
			idx = self.coeffs.index(expr)
			return [1,idx-1]

		return [-2]



class Node:

	def __init__(self, key=0, value=None):
		self.map = {}
		self.key = key
		self.value = value


	def get_key(self):
		return self.key

	def get_value(self):
		return self.value

	def insert(self,iis, value):
		# print "  processing: ", iis
		# print "    ", self.key, self.map
		if len(iis) > 1:
			ii = iis[0]
			if ii not in self.map:
				# print "  new node for key: ", ii
				self.map[ii] = Node(key=ii)
			return self.map[ii].insert(iis[1:], value)
		if len(iis) == 1:
			ii = iis[0]
			if ii not in self.map:
				# print "  new node for key: ", ii
				self.map[ii] = Node(key=ii, value=value)
				return True
			else:
				return False
		return False

	def lookup(self,iis):
		if len(iis) > 1:
			ii = iis[0]
			if ii in self.map:
				return self.map[ii].lookup(iis[1:])
			else:
				return False, None
		if len(iis) == 1:
			ii = iis[0]
			if ii in self.map:
				return True, self.map[ii].get_value()
			else:
				return False, None
		return False, None

